/*
    Copyright (C) 1996-2000  Ulric Eriksson <ulric@siag.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>
#include <sys/stat.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/xpm.h>
#include "../common/common.h"
#include "../common/cmalloc.h"
#include "filesel.h"

static pid_t splashpid;

/* ---
Get the text in a label.
*/

String label_get(Widget w)
{
	String p = NULL;
	XtVaGetValues(w, XtNlabel, &p, (char *)0);
	return p;
}

/* ---
Set the text in a label.
*/

void label_set(Widget w, String p)
{
        char *q, *r;
        int must_set = 0;

        if (w == None || p == NULL) return;
        q = _((char *)p);
        if (q == NULL) return;
        r = label_get(w);

        if (r == NULL) {
                must_set = 1;
        } else {
                must_set = (strcmp(q, r) != 0);
        }

        if (must_set) XtVaSetValues(w, XtNlabel, q, (char *)0);
}

/* ---
Get the state of a toggle widget
*/

Boolean state_get(Widget w)
{
	Boolean i = False;
	if (w != None)
		XtVaGetValues(w, XtNstate, &i, (char *)0);
	return i;
}

/* this is needed if we include vanilla Athena headers */
#ifndef XtNshadowWidth
#define XtNshadowWidth "shadowWidth"
#endif

/* ---
If no widget, do nothing.
If same state and same shadow width as before, do nothing.
If set, use shadow width ssw.
If unset, use shadow width usw.
*/

void state_set(Widget w, Boolean i, int ssw, int usw)
{
	Boolean set;
	Dimension osw, nsw;
	if (w == None) return;
	XtVaGetValues(w,
		XtNstate, &set,
		XtNshadowWidth, &osw,
		(char *)0);
	if (i) nsw = ssw;
	else nsw = usw;
	if (i == set && nsw == osw) return;
	XtVaSetValues(w,
		XtNstate, i,
		XtNshadowWidth, nsw,
		(char *)0);
}

/* ---
Get the width of a widget.
*/

Dimension width_get(Widget w)
{
	Dimension x = 80;
	if (w != None)
		XtVaGetValues(w, XtNwidth, &x, (char *)0);
	return x;
}

/* ---
Get the height of a widget.
*/

Dimension height_get(Widget w)
{
	Dimension x = 20;
	if (w != None)
		XtVaGetValues(w, XtNheight, &x, (char *)0);
	return x;
}

void start_splash(void)
{
	char *p = getenv("SIAG_SPLASH");

	if (p) {
		splashpid = fork();

		if (splashpid == 0) {
			char cmd[1024];
			char image[1024];
			sprintf(cmd, "%s/plugins/image", libdir);
			sprintf(image, "%s/common/bitmaps/SO3.xpm", datadir);
			execlp(cmd, "image", image, (char *)NULL);
			_exit(0);
		}
	}
}

void stop_splash(void)
{
	if (splashpid > 0) kill(splashpid, SIGINT);
}

Pixmap load_pixmap(Display *dpy, Pixel color, char *pm)
{
	XpmAttributes xa;
	Pixmap pm_return;
	int result;
	XpmColorSymbol symbol;
	char fn[1024];
	char *pixpath = getenv("PIXPATH");
	struct stat statbuf;
	int found = 0;
	Screen *scr = DefaultScreenOfDisplay(dpy);

	if (pm[0] == '/') {
		strcpy(fn, pm);
		found = 1;
	} else if (pixpath) {
		sprintf(fn, "%s/%s", pixpath, pm);
		found = !stat(fn, &statbuf);
	}
	if (!found) sprintf(fn, "%s/common/bitmaps/kde/%s", datadir, pm);
	found = !stat(fn, &statbuf);
	if (!found) sprintf(fn, "%s/common/bitmaps/%s", datadir, pm);

#if 0
	XtVaGetValues(w, XtNbackground, &color, (char *)0);
#endif
	xa.closeness = 40000;
	xa.exactColors = False;
	xa.valuemask = XpmCloseness | XpmExactColors;

#if 1
	symbol.name = NULL;
	symbol.value = "none";
	symbol.pixel = color;
	xa.colorsymbols = &symbol;
	xa.numsymbols = 1;
	xa.valuemask |= XpmColorSymbols;
#endif
	result = XpmReadFileToPixmap(dpy,
		XRootWindowOfScreen(scr),
		fn, &pm_return, NULL, &xa);

	if (result != XpmSuccess) {
		fprintf(stderr, "XpmReadFileToPixmap(%s) returns %s\n",
			fn, XpmGetErrorString(result));
		return None;
	}
	return pm_return;
}

/* color allocation stuff snarfed from the xmagick library */
#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif

static int rgb_shift = 10;

typedef struct xmagick_color {
	XColor color;
	unsigned long rgb;
} xmagick_color;

static xmagick_color known_colors[128];
static int ncolors, maxcolors = 128, lastcolor;

void free_colors(void)
{
	free(known_colors);
/*	known_colors = NULL;
*/
	ncolors = 0;
	lastcolor = -1;
}

void alloc_color(Display *display, Colormap colormap, XColor *color)
{
	int i, j, lower, upper;
	int cr, cg, cb, d;
	unsigned long rgb;
	cr = color->red >> rgb_shift;
	cg = color->green >> rgb_shift;
	cb = color->blue >> rgb_shift;
	rgb = (cr << 16) + (cg << 8) + cb;

	if (lastcolor != -1 && known_colors[lastcolor].rgb == rgb) {
		color->pixel = known_colors[lastcolor].color.pixel;
		return;
	}
	lower = 0;
	upper = ncolors-1;
	while (lower <= upper) {
		i = (lower+upper)/2;
		d = rgb-known_colors[i].rgb;
		if (d == 0) {
			lastcolor = i;
			color->pixel = known_colors[i].color.pixel;
			return;
		} else if (d > 0) {
			lower = i+1;
		} else {
			upper = i-1;
		}
	}
	if (ncolors >= maxcolors) {
		color->pixel = 0;
		return;
	}
	color->flags = DoRed|DoGreen|DoBlue;
	if (XAllocColor(display, colormap, color) == 0) {
		int dr, dg, db, bd = INT_MAX, best = 0;
		for (j = 0; j < ncolors; j++) {
			XColor c = known_colors[j].color;
			dr = cr - (c.red >> rgb_shift);
			dg = cg - (c.green >> rgb_shift);
			db = cb - (c.blue >> rgb_shift);
			d = dr*dr+dg*dg+db*db;
			if (d < bd) {
				bd = d;
				best = j;
			}
		}
		color->pixel = known_colors[best].color.pixel;
	}
	if (known_colors[upper].rgb > rgb) i = upper;
	else i = lower;
	if (i < 0) i = 0;
	if (i > ncolors) i = ncolors;
	for (j = ncolors; j > i; j--) {
		known_colors[j] = known_colors[j-1];
	}
	known_colors[i].color = *color;
	known_colors[i].rgb = rgb;
	known_colors[i].color.pixel = color->pixel;
	lastcolor = i;
	ncolors++;
}

void query_color(Display *display, Colormap colormap, XColor *color)
{
	int i;

	for (i = 0; i < ncolors; i++) {
		if (known_colors[i].color.pixel == color->pixel) {
			color->red = known_colors[i].color.red;
			color->green = known_colors[i].color.green;
			color->blue = known_colors[i].color.blue;
			return;
		}
	}
	if (i >= maxcolors) {
		color->red = 0;
		color->green = 0;
		color->blue = 0;
		return;
	}
	XQueryColor(display, colormap, color);
	known_colors[i].color = *color;
	ncolors++;
}

int xerror_handler(Display *dpy, XErrorEvent *event)
{
	char b[1000];

	XGetErrorText(dpy, event->error_code, b, 999);
	puts(b);
	return 0;
}

void theme_init(Display *display)
{
        char b[1024], *tr;
        XrmDatabase themedb, olddb;
        XrmValue vr;
 
        sprintf(b, "%s/theme", siag_basedir);
        themedb = XrmGetFileDatabase(b);
        if (themedb) {
                olddb = XrmGetDatabase(display);
                XrmMergeDatabases(themedb, &olddb);
                XrmSetDatabase(display, olddb);
                if (XrmGetResource(olddb, "PIXPATH", XtNstring, &tr, &vr)) {
			if (!strncmp(vr.addr, ".../", 4)) {
				sprintf(b, "PIXPATH=%s%s", datadir, vr.addr+3);
			} else {
				sprintf(b, "PIXPATH=%s", vr.addr);
			}
                        putenv(cstrdup(b));
                }
                if (XrmGetResource(olddb, "XAWM_THEME", XtNstring, &tr, &vr)) {
                        sprintf(b, "XAWM_THEME=%s", vr.addr);
                        putenv(cstrdup(b));
                }
        }
}

static void copy_file(char *from, char *to)
{
	FILE *fp1, *fp2;
	int c;

	remove(to);
	fp1 = fopen(from, "r");
	fp2 = fopen(to, "w");
	if (!fp1 || !fp2) return;
	while ((c = fgetc(fp1)) != EOF) fputc(c, fp2);
	fclose(fp1);
	fclose(fp2);
}

int select_theme(Widget pw)
{
	char path[1024], name[1024], fn[1024];
	char fmt[80];
	char extra[1024];
	int n;

	sprintf(extra, "Home=%s", getenv("HOME"));
	sprintf(path, "%s/common/themes", datadir);
	name[0] = '\0';
	n = fsel_input(pw, path, name, NULL, fmt, extra, 0);
	if (!n) return 0;
	sprintf(fn, "%s/%s", path, name);
	sprintf(extra, "%s/.siag/theme", getenv("HOME"));
	copy_file(fn, extra);
	return 1;
}

